{% extends "base.html" %}

{% block relbar1 %}
{{ super() }}
<div class="deck">
<div class="header">
    {% if release_version %}
        <p>
        Documentation for pulsar {{ version }}. For development docs,
        <a href="http://quantmind.github.io/pulsar/{{ pagename }}{{ file_suffix }}">go here</a>.
        </p>
    {% else %}
        <p class="developmentversion">
        Documentation for pulsar's DEVELOPMENT version. Get the
        <a href="http://pythonhosted.org/pulsar/{{ pagename }}{{ file_suffix }}">release docs here</a>.
        </p>
    {% endif %}
</div>
</div>
{% endblock %}
{% block relbar2 %}{% endblock %}
