.panel {
    background: #fff;
    border: solid 1px #E6E6E6;
}

.main {
    padding: 60px 0;
}

#users {
    height: 400px;
}
#messages {
    height: 100%;
    min-height: 620px;
}
#messages p {
    padding: 10px 5px;
    margin: 0;
    color: #666666;
    border-bottom: 1px solid #F2F2F2;
}

#input textarea {
    border: solid 1px #E6E6E6;
}
.layout {
    margin: 20px auto;
}
