<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="utf-8"/>
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>pulsar-django chat application</title>
        <link href="http://netdna.bootstrapcdn.com/bootswatch/2.3.1/readable/bootstrap.min.css" rel="stylesheet">
        <link href="static/djangoapp/chat.css" rel="stylesheet">
    </head>
    <body>
        {% block content %}{% endblock %}
    <script src="//code.jquery.com/jquery-2.0.3.min.js"></script>
    <script src="//cdnjs.cloudflare.com/ajax/libs/json3/3.2.6/json3.min.js"></script>
    <script src="/static/djangoapp/chat.js"></script>
    {% block script %}{% endblock %}
    </body>
</html>
